import { beforeEach, expect, test, mock } from "bun:test";
import { reset, expectCalls, getWidgets } from "./test/widget-mock.ts";

const { Progress, withProgress } = await import("@paperclover/console/Progress");
type Progress<T extends Record<string, unknown> = Record<never, unknown>> = import("@paperclover/console/Progress").Progress<T>;

beforeEach(reset);

test("default options and basic rendering", () => {
    const progress = new Progress("Loading");

    expect(progress.text).toEqual("Loading");
    expect(progress.beforeText).toEqual("");
    expect(progress.value).toEqual(0);
    expect(progress.total).toEqual(100);

    // Format should include a progress bar
    const formatted = progress.format(0);
    expect(formatted).toContain("Loading");
    // The progress bar might not contain "█" at 0% progress,
    // but it should contain color codes and spaces for the bar
    expect(formatted).toContain("\u001b[38;2;"); // Color code
});

test("progress percentage displays correctly", () => {
    const progress = new Progress("Loading");

    // Test zero progress
    progress.value = 0;
    expect(progress.format(0)).toContain("Loading");

    // Test partial progress
    progress.value = 50;
    const halfProgress = progress.format(0);
    expect(halfProgress).toContain("Loading");

    // Test full progress
    progress.value = 100;
    const fullProgress = progress.format(0);
    expect(fullProgress).toContain("Loading");

    // Should be visibly different
    expect(progress.format(0)).not.toEqual(halfProgress);
});

test("custom options apply correctly", () => {
    const progress = new Progress({
        text: "Custom Progress",
        beforeText: "Loading:",
        barWidth: 20,
        barStyle: "ascii",
        value: 25,
        total: 200
    });

    expect(progress.text).toEqual("Custom Progress");
    expect(progress.beforeText).toEqual("Loading:");
    expect(progress.value).toEqual(25);
    expect(progress.total).toEqual(200);

    const formatted = progress.format(0);
    expect(formatted).toContain("Custom Progress");
    expect(formatted).toContain("Loading:");
    // With ascii style, it should contain these characters
    expect(formatted).toContain("[");
    expect(formatted).toContain("]");
});

test("spinner options", () => {
    // Test with custom spinner
    const customSpinnerProgress = new Progress({
        text: "With spinner",
        spinner: {
            frames: ["A", "B", "C"],
            fps: 10,
            color: "red"
        }
    });

    expect(customSpinnerProgress.fps).toBeGreaterThan(0); // Should have fps when spinner is enabled
    const formatted = customSpinnerProgress.format(0);
    expect(formatted).toContain("A"); // Should include the first spinner frame

    // Test with spinner disabled
    const noSpinnerProgress = new Progress({
        text: "No spinner",
        spinner: null
    });

    expect(noSpinnerProgress.fps).toBe(0); // Should have 0 fps when spinner is disabled
    const noSpinnerFormatted = noSpinnerProgress.format(0);
    expect(noSpinnerFormatted).not.toContain("A");
    expect(noSpinnerFormatted).not.toContain("B");
    expect(noSpinnerFormatted).not.toContain("C");
});

test("text getter and setter", () => {
    const progress = new Progress("Initial text");
    expect(progress.text).toEqual("Initial text");

    progress.text = "Updated text";
    expect(progress.text).toEqual("Updated text");

    expectCalls(progress, {
        redraws: 1
    });
});

test("beforeText getter and setter", () => {
    const progress = new Progress({
        text: "Main text",
        beforeText: "Initial before"
    });
    expect(progress.beforeText).toEqual("Initial before");

    progress.beforeText = "Updated before";
    expect(progress.beforeText).toEqual("Updated before");

    expectCalls(progress, {
        redraws: 1
    });
});

test("value and total getters and setters", () => {
    const progress = new Progress("Progress");
    expect(progress.value).toEqual(0);
    expect(progress.total).toEqual(100);

    progress.value = 25;
    expect(progress.value).toEqual(25);
    expectCalls(progress, {
        redraws: 1
    });

    progress.total = 200;
    expect(progress.total).toEqual(200);
    expectCalls(progress, {
        redraws: 2
    });
});

test("function for text field and props", () => {
    const textFn = (props: { value: number; total: number; progress: number; customProp: string }) =>
        `${props.customProp}: ${props.value}/${props.total} (${Math.round(props.progress * 100)}%)`;

    const progress = new Progress<{ customProp: string }>({
        text: textFn,
        props: { customProp: "Loading" },
        value: 30,
        total: 120
    });

    // Text function should use props and calculate progress
    expect(progress.text).toEqual("Loading: 30/120 (25%)");

    // Updating props should reflect in text
    progress.props = { customProp: "Downloading" };
    expect(progress.text).toEqual("Downloading: 30/120 (25%)");

    // Updating value should reflect in text
    progress.value = 60;
    expect(progress.text).toEqual("Downloading: 60/120 (50%)");

    expectCalls(progress, {
        redraws: 2
    });
});

test("update method and its overloads", () => {
    const progress = new Progress<{ status: string }>({
        text: (props) => `${props.status}: ${props.value}/${props.total}`,
        props: { status: "Loading" }
    });

    // Initial state
    expect(progress.text).toEqual("Loading: 0/100");

    // Update with just value
    progress.update(50);
    expect(progress.value).toEqual(50);
    expect(progress.text).toEqual("Loading: 50/100");

    // Update with value and props
    progress.update(75, { status: "Downloading" });
    expect(progress.value).toEqual(75);
    expect(progress.text).toEqual("Downloading: 75/100");

    expectCalls(progress, {
        redraws: 2
    });
});

test("success and error methods", () => {
    const progress = new Progress("Working");

    // Test success with default message
    progress.success();
    expectCalls(progress, {
        successCalls: ["Working"]
    });

    // Test success with custom message
    const progress2 = new Progress("Working");
    progress2.success("Completed");
    expectCalls(progress2, {
        successCalls: ["Completed"]
    });

    // Test error with default message
    const progress3 = new Progress("Working");
    progress3.error();
    expectCalls(progress3, {
        errorCalls: ["Working"]
    });

    // Test error with custom message
    const progress4 = new Progress("Working");
    progress4.error("Failed");
    expectCalls(progress4, {
        errorCalls: ["Failed"]
    });

    // Test error with Error object
    const error = new Error("Something went wrong");
    const progress5 = new Progress("Working");
    progress5.error(error);
    expectCalls(progress5, {
        errorCalls: [error]
    });
});

test("withProgress function", async () => {
    const expectedResult = Symbol("result");
    let bar!: Progress;

    try {
        // We need to use try/finally to ensure we can access bar for validation
        await withProgress("Processing", async (progress) => {
            bar = progress;
            expect(progress.text).toEqual("Processing");

            // Update progress during operation
            progress.value = 50;

            return expectedResult;
        });
    } finally {
        expectCalls(bar, {
            successCalls: ["Completed"],
            redraws: 1
        });
    }
});

test("withProgress function with custom success text", async () => {
    let bar!: Progress;

    try {
        await withProgress({
            text: "Processing",
            successText: "Task completed successfully"
        }, async (progress) => {
            bar = progress;
            return "data";
        });
    } finally {
        expectCalls(bar, {
            successCalls: ["Task completed successfully"],
            redraws: 0
        });
    }
});

test("withProgress function with success text function", async () => {
    let bar!: Progress;
    const data = { count: 42 };

    try {
        await withProgress({
            text: "Processing",
            successText: (result: { count: number }) => `Processed ${result.count} items`
        }, async (progress) => {
            bar = progress;
            return data;
        });
    } finally {
        expectCalls(bar, {
            successCalls: ["Processed 42 items"],
            redraws: 0
        });
    }
});

test("withProgress function with error", async () => {
    const error = new Error("Process failed");
    let bar!: Progress;

    try {
        await withProgress({
            text: "Processing",
            failureText: "Task failed"
        }, async (progress) => {
            bar = progress;
            throw error;
        });
        expect.unreachable();
    } catch (e) {
        expect(e).toEqual(error);
        expectCalls(bar, {
            errorCalls: ["Task failed"],
            redraws: 0
        });
    }
});

test("withProgress function with error text function", async () => {
    const error = new Error("Process failed with code 500");
    let bar!: Progress;

    try {
        await withProgress({
            text: "Processing",
            failureText: (err) => `Error: ${err.message}`
        }, async (progress) => {
            bar = progress;
            throw error;
        });
        expect.unreachable();
    } catch (e) {
        expect(e).toEqual(error);
        expectCalls(bar, {
            errorCalls: ["Error: Process failed with code 500"],
            redraws: 0
        });
    }
});