import { Widget } from './Widget.ts';
import { getColor, type CustomLoggerColor } from './internal.ts';

export interface SpinnerOptions<Props extends Record<string, unknown>> {
  /** Text displayed to the right of the spinner. */
  text: string | ((props: Props) => string);
  /** Color of the spinner. */
  color?: CustomLoggerColor | false;
  /** Sequence of frames for the spinner. */
  frames?: readonly string[];
  /** Frames per second of the Spinner. */
  fps?: number;
  /** Properties to be passed to the `text` formatting function. */
  props?: Props;
}

export const defaultSpinnerOptions = {
  text: 'Loading...',
  color: 'blueBright',
  frames: ['⠋', '⠙', '⠹', '⠸', '⠼', '⠴', '⠦', '⠧', '⠇', '⠏'],
  fps: 12.5,
} as const;

export class Spinner<Props extends Record<string, unknown> = Record<string, unknown>> extends Widget {
  readonly frames: readonly string[];
  #text: string | ((props: Props) => string);
  #color: ((text: string) => string) | null;
  #props: Props;
  fps: number;

  constructor(options: SpinnerOptions<Props> | string) {
    super();
    if (typeof options === 'string') {
      options = { text: options };
    }
    this.#text = options.text ?? defaultSpinnerOptions.text;
    const color = options.color ?? defaultSpinnerOptions.color;
    this.#color = color === false ? null : getColor(color);
    this.frames = options.frames ?? defaultSpinnerOptions.frames;
    this.fps = options.fps ?? defaultSpinnerOptions.fps;
    this.#props = options.props ?? ({} as Props);
  }

  /** Text displayed to the right of the spinner. */
  get text(): string {
    return typeof this.#text === 'function' ? this.#text(this.#props) : this.#text;
  }

  set text(value: string | (() => string)) {
    this.#text = value;
    this.redraw();
  }

  /** Properties to be passed to `text` and `beforeText` formatting functions. */
  set props(value: Partial<Props>) {
    this.#props = {
      ...this.#props,
      ...value,
    };
    this.redraw();
  }

  get props(): Props {
    return {
      ...this.#props,
    };
  }

  /**
   * Updates the spinner by supplying either a new `message` string or a partial object of props to
   * be used by the custom message function.
   */
  update(newProps: Partial<Props>): void;
  update(newMessage: string): void;
  update(newData: string | Partial<Props>) {
    if (typeof newData === 'string') {
      this.text = newData;
    } else {
      this.#props = { ...this.#props, ...newData };
      this.redraw();
    }
  }

  format(now: number): string {
    const frame = Math.floor(now / (1000 / this.fps)) % this.frames.length;
    const frameText = this.frames[frame]!;
    return (
      (this.#color ? this.#color(frameText) : frameText) +
      ' ' +
      this.text
    );
  }

  success(message?: string): void {
    super.success(message ?? this.text);
  }

  error(message?: string | Error): void {
    super.error(message ?? this.text);
  }
}

export interface WithSpinnerOptions<Props extends Record<string, unknown>, T>
  extends SpinnerOptions<Props> {
  successText?: string | ((result: T) => string);
  failureText?: string | ((error: Error) => string);
}

/** Calls a function with a spinner. */
export async function withSpinner<Props extends Record<string, unknown>, T>(
  spinnerOptions: WithSpinnerOptions<Props, T> | string,
  fn: (spinner: Spinner<Props>) => Promise<T>
): Promise<T>;
export async function withSpinner(opts: any, fn: any) {
  const spinner = new Spinner(opts);

  try {
    const result = await fn(spinner);
    if (spinner.active) {
      spinner.success(
        opts.successText
          ? typeof opts.successText === 'function'
            ? opts.successText(result)
            : opts.successText
          : opts.text
            ? typeof opts.text === 'function'
              ? opts.text(spinner.props)
              : opts.text
            : 'Completed'
      );
    }
    return result;
  } catch (error: any) {
    spinner.error(
      typeof opts.failureText === 'function' ? opts.failureText(error) : opts.failureText ?? error
    );
    throw error;
  }
}
