export const isUnicodeSupported =
    process.platform === 'win32'
        ? Boolean(process.env.CI) ||
        Boolean(process.env.WT_SESSION) || // Windows Terminal
        process.env.ConEmuTask === '{cmd::Cmder}' || // ConEmu and cmder
        process.env.TERM_PROGRAM === 'vscode' ||
        process.env.TERM === 'xterm-256color' ||
        process.env.TERM === 'alacritty' ||
        process.env.TERMINAL_EMULATOR === 'JetBrains-JediTerm'
        : process.env.TERM !== 'linux';

export const errorSymbol = isUnicodeSupported ? '✖' : 'x';
export const successSymbol = isUnicodeSupported ? '✔' : '√';
export const infoSymbol = isUnicodeSupported ? 'ℹ' : 'i';
export const warningSymbol = isUnicodeSupported ? '⚠' : '‼';

let filters: string[] = [];
let filterGeneration = 0;
export function setLogFilter(...newFilters: Array<string | string[]>) {
    filters = newFilters.flat().map(filter => filter.toLowerCase());
    filterGeneration++;
}

export function isLogVisible(id: string, defaultVisibility = true) {
    for (const filter of filters) {
        if (filter === '*') {
            defaultVisibility = true;
        } else if (filter === '-*') {
            defaultVisibility = false;
        } else if (filter === id || id.startsWith(filter + ':')) {
            defaultVisibility = true;
        } else if (filter === '-' + id || id.startsWith('-' + filter + ':')) {
            defaultVisibility = false;
        }
    }
    return defaultVisibility;
}

if (process.env.DEBUG !== undefined) {
    setLogFilter(
        String(process.env.DEBUG)
            .split(',')
            .map(x => x.trim())
            .map(x => (['1', 'true', 'all'].includes(x.toLowerCase()) ? '*' : x))
    );
}


/** Taken from https://github.com/debug-js/debug/blob/d1616622e4d404863c5a98443f755b4006e971dc/src/node.js#L35. */
const debugColors = [
    20, 21, 26, 27, 32, 33, 38, 39, 40, 41, 42, 43, 44, 45, 56, 57, 62, 63, 68, 69, 74, 75, 76, 77,
    78, 79, 80, 81, 92, 93, 98, 99, 112, 113, 128, 129, 134, 135, 148, 149, 160, 161, 162, 163, 164,
    165, 166, 167, 168, 169, 170, 171, 172, 173, 178, 179, 184, 185, 196, 197, 198, 199, 200, 201,
    202, 203, 204, 205, 206, 207, 208, 209, 214, 215, 220, 221,
];

/** Converts non string objects into a string the way Node.js' console.log does it. */
function stringify(...data: any[]) {
    return data
        .map(obj => {
            if (typeof obj === 'string') {
                return obj;
            } else if (obj instanceof Error) {
                return formatErrorObj(obj);
            }
            return inspect(obj, false, 4, true);
        })
        .join(' ');
}

/**
 * Selects a color for a debug namespace.
 *
 * Taken from https://github.com/debug-js/debug/blob/master/src/common.js.
 */
function selectColor(namespace: string) {
    let hash = 0;

    for (let i = 0; i < namespace.length; i++) {
        hash = (hash << 5) - hash + namespace.charCodeAt(i);
        hash |= 0; // Convert to 32bit integer
    }

    return debugColors[Math.abs(hash) % debugColors.length]!;
}



const formatImplementation = {
    's': (data: StringLike) => String(data),
    'd': (data: number) => String(data),
    'i': (data: number) => String(Math.floor(data)),
    'f': (data: number) => String(data),
    'x': (data: number) => data.toString(16),
    'X': (data: number) => data.toString(16).toUpperCase(),
    'o': (data: any) => JSON.stringify(data),
    'O': (data: any) => JSON.stringify(data, null, 2),
    'c': () => '',
    'j': (data: any) => JSON.stringify(data),
};

function format(fmt: any, ...args: any[]) {
    if (typeof fmt === 'string') {
        let index = 0;
        const result = fmt.replace(/%[%sdifoxXcj]/g, match => {
            if (match === '%%') {
                return '%';
            }
            const arg = args[index++];
            return (formatImplementation as any)[match[1]!](arg);
        });

        if (index === 0 && args.length > 0) {
            return result + ' ' + stringify(...args);
        }

        return result;
    }
    return stringify(fmt, ...args);
}

const LogFunction = {
    __proto__: Function.prototype,
    [Symbol.for('nodejs.util.inspect.custom')](depth: number, options: any) {
        return options.stylize(`[LogFunction: ${(this as any).name}]`, 'special');
    },
};

/**
 * Creates a logger function with a pseudo-random color based off the namespace.
 *
 * A custom color can be assigned by doing any of the following:
 *
 * - Passing a color argument with a color name "blue"
 * - Passing a color argument with a hex value "#0000FF"
 * - Passing a color argument with an ANSI 256 palette value (0-255)
 * - Passing a color argument with a RGB value [0, 0, 255]
 * - Using chalk or another formatter on the namespace name.
 */
export function scoped(
    name: string,
    opts: CustomLoggerOptions | CustomLoggerColor = {}
): LogFunction {
    if (typeof opts === 'string' || Array.isArray(opts) || typeof opts === 'number') {
        opts = { color: opts };
    }
    const {
        id = name,
        color = undefined,
        coloredText = false,
        boldText = false,
        debug = false,
    } = opts;
    const strippedName = stripAnsi(name);
    const colorFn = name.includes('\x1b')
        ? chalk
        : color
            ? getColor(color)
            : chalk.ansi256(selectColor(name));
    const coloredName = colorFn.bold(name);
    const fn = ((fmt: unknown, ...args: any[]) => {
        if (!fn.visible) {
            return;
        }

        const data = format(fmt, ...args).replace(/\n/g, '\n ' + ' '.repeat(strippedName.length));

        if (fmt === undefined && args.length === 0) {
            writeLine();
        } else {
            writeLine(
                coloredName + ' ' + (coloredText ? (boldText ? colorFn.bold(data) : colorFn(data)) : data)
            );
        }
    }) as LogFunction;
    Object.setPrototypeOf(fn, LogFunction);
    Object.defineProperty(fn, 'name', { value: id });
    let gen = filterGeneration;
    let visible = isLogVisible(id, !debug);
    Object.defineProperty(fn, 'visible', {
        get: () => {
            if (gen !== filterGeneration) {
                gen = filterGeneration;
                visible = isLogVisible(id, !debug);
            }
            return visible;
        }
    });
    return fn;
}


/** Built in blue "info" logger. */
export const info = /* @__PURE__ */ scoped('info', {
    color: 'blueBright',
});

/** Built in yellow "warn" logger. */
export const warn = /* @__PURE__ */ scoped('warn', {
    color: 'yellowBright',
    coloredText: true,
});

const _trace = /* @__PURE__ */ scoped('trace', {
    color: 208,
});

/** Built in orange "trace" logger. Prints a stack trace after the message. */
export const trace = /* @__PURE__ */ ((trace: any) => (Object.defineProperty(trace, 'visible', { get: () => _trace.visible }), trace))(function trace(...data: any[]) {
    if (_trace.visible) {
        _trace(...(data.length === 0 ? [' '] : data));
        writeLine(formatStackTrace(new Error()).split('\n').slice(1).join('\n'));
    }
}) as typeof _trace;

/** Built in red "error" logger, uses a unicode X instead of the word Error. */
export const error = /* @__PURE__ */ scoped(errorSymbol, {
    id: 'error',
    color: 'redBright',
    coloredText: true,
    boldText: true,
});

/** Built in cyan "debug" logger. */
export const debug = scoped('debug', {
    color: 'cyanBright',
    debug: true,
});

/** Built in green "success" logger, uses a unicode Check instead of the word Success. */
export const success = /* @__PURE__ */ scoped(successSymbol, {
    id: 'success',
    color: 'greenBright',
    coloredText: true,
    boldText: true,
});

import chalk, { type ChalkInstance } from 'chalk';
import { inspect } from 'util';
import { type LogFunction, type CustomLoggerColor, type CustomLoggerOptions, type StringLike, getColor } from './internal.ts';
import { formatErrorObj, formatStackTrace } from './error.ts';
import stripAnsi from 'strip-ansi';
import { writeLine } from './Widget.ts';

export { writeLine } from './Widget.ts';
