import { expect, test } from "bun:test";
import { formatErrorObj, formatStackTrace, CLIError, platformSimplifyErrorPath } from "./error.ts";
import chalk from "chalk";

test("formatErrorObj formats basic errors", () => {
  const error = new Error("Something went wrong");
  const formatted = formatErrorObj(error);

  // Basic error formatting should contain the error name and message
  expect(formatted).toContain("Error: Something went wrong");
  // Should contain a stack trace
  expect(formatted).toContain("at ");
});

test("formatErrorObj with PrintableError interface", () => {
  const printableError = new Error("Custom message") as any;
  printableError.name = "CustomError";
  printableError.description = "This is a detailed description of the error";
  
  const formatted = formatErrorObj(printableError);
  
  // Should contain name, message and description
  expect(formatted).toContain("CustomError: Custom message");
  expect(formatted).toContain("This is a detailed description of the error");
  expect(formatted).toContain("at ");
});

test("formatErrorObj with hideName option", () => {
  const printableError = new Error("Just the message") as any;
  printableError.hideName = true;
  
  const formatted = formatErrorObj(printableError);
  
  // Should not contain "Error:"
  expect(formatted).not.toContain("Error:");
  // Don't check the exact output since the stack trace might vary
});

test("formatErrorObj with hideStack option", () => {
  const printableError = new Error("No stack trace") as any;
  printableError.hideStack = true;
  
  const formatted = formatErrorObj(printableError);
  
  // Should contain message but not stack trace
  expect(formatted).toContain("Error: No stack trace");
  expect(formatted).not.toContain("at <anonymous>");
});

test("formatStackTrace with V8 style stack", () => {
  const error = new Error("Stack trace test");
  const formatted = formatStackTrace(error);
  
  // Should format the stack trace
  expect(formatted).toContain("at ");
  // May not have color codes in test environment
});

test("formatStackTrace with empty stack", () => {
  const error = new Error("No stack");
  error.stack = undefined;
  
  const formatted = formatStackTrace(error);
  
  // Should handle undefined stack
  expect(formatted).toEqual("");
});

test("platformSimplifyErrorPath with cwd", () => {
  const cwd = process.cwd();
  const fullPath = `${cwd}/src/error.ts`;
  const simplified = platformSimplifyErrorPath(fullPath);
  
  // Should convert to a relative path
  expect(simplified).toEqual("./src/error.ts");
});

test("platformSimplifyErrorPath with external path", () => {
  const externalPath = "/usr/local/lib/node_modules/module.js";
  const simplified = platformSimplifyErrorPath(externalPath);
  
  // Should keep external paths unchanged
  expect(simplified).toEqual(externalPath);
});

test("CLIError class", () => {
  const cliError = new CLIError(
    "Command failed",
    "Try running with --help for available options"
  );
  
  expect(cliError.message).toEqual("Command failed");
  expect(cliError.description).toEqual("Try running with --help for available options");
  expect(cliError.hideStack).toEqual(true);
  expect(cliError.hideName).toEqual(true);
  
  const formatted = formatErrorObj(cliError);
  
  // Should format according to the PrintableError interface
  expect(formatted).toEqual("Command failed\nTry running with --help for available options\n");
  expect(formatted).not.toContain("CLIError");
  expect(formatted).not.toContain("at ");
});

test("pretty printing of objects in stack traces", () => {
  try {
    // Create an error with a complex object
    const complexObj = { a: 1, b: { c: 2 }, d: [1, 2, 3] };
    JSON.parse("{invalid}");
  } catch (error) {
    const formatted = formatErrorObj(error as Error);
    
    // Should format the error message
    expect(formatted).toContain("SyntaxError");
    expect(formatted).toContain("at ");
  }
});