import { beforeEach, expect, test, mock } from "bun:test";
import { spawn } from "child_process";

// We need to use a subprocess to test injection since it happens immediately on import
// and would affect our test environment

function runSubprocess(testType: string, env: Record<string, string> = {}): Promise<string> {
  return new Promise((resolve, reject) => {
    // Merge the current process.env with any additional env variables
    const environment = { ...process.env, ...env };
    
    const subprocess = spawn("bun", ["src/test/subprocess.ts", testType], {
      env: environment
    });
    
    let stdout = "";
    let stderr = "";
    
    subprocess.stdout.on("data", (data) => {
      stdout += data.toString();
    });
    
    subprocess.stderr.on("data", (data) => {
      stderr += data.toString();
    });
    
    subprocess.on("error", (error) => {
      reject(error);
    });
    
    subprocess.on("close", (code) => {
      if (code !== 0) {
        reject(new Error(`Subprocess exited with code ${code}: ${stderr}`));
      } else {
        resolve(stdout + stderr); // Combine stdout and stderr since our console outputs go to stderr
      }
    });
  });
}

test("basic injection is applied correctly", async () => {
  // Enable debug logs with DEBUG=1 and force color with FORCE_COLOR
  const output = await runSubprocess("basic-injection", { 
    DEBUG: "1",
    FORCE_COLOR: "1"
  });
  
  // Check that console methods are correctly injected
  expect(output).toContain("INJECTED_LOG");
  expect(output).toContain("INJECTED_INFO");
  expect(output).toContain("INJECTED_WARN");
  expect(output).toContain("INJECTED_ERROR");
  expect(output).toContain("INJECTED_DEBUG");
  
  // Assert should only output when condition is false
  expect(output).toContain("INJECTED_ASSERT");
  expect(output).not.toContain("This should not appear");
});

test("timer functions work correctly", async () => {
  const output = await runSubprocess("time-functions");
  
  // Check timer functions
  expect(output).toContain("TIMER_LABEL");
  expect(output).toContain("ms]"); // Time output should include milliseconds
  
  // Warning for non-existent timer
  expect(output).toContain("does not exist");
});

test("count functions work correctly", async () => {
  const output = await runSubprocess("count-functions");
  
  // Check counter functions
  expect(output).toContain("COUNTER_LABEL: 1");
  expect(output).toContain("COUNTER_LABEL: 2");
  expect(output).toContain("COUNTER_LABEL: 1"); // After reset
});

test("environment variables affect injection", async () => {
  const output = await runSubprocess("disabled-injection", { NO_COLOR: "1" });
  
  // When NO_COLOR is set, output should still have the message
  expect(output).toContain("DISABLED_COLOR_LOG");
  // Note: Chalk may still output color codes even with NO_COLOR=1 in some environments
  // So we're not checking for absence of color codes here
});