import chalk from 'chalk';
import { debug, error, info, trace, warn } from './console.ts';
import { Spinner } from './Spinner.ts';
import { errorAllWidgets, writeLine } from './Widget.ts';

// Basic Logging Functions
console.log = info;
console.info = info;
console.warn = warn;
console.error = error;
console.debug = debug;

// Assert
console.assert = (condition, ...msg: Parameters<typeof error>) => {
  if (!condition) {
    error(...msg);
  }
};

// Time
const timers = new Map<string, { start: number; spinner: Spinner }>();
console.time = (label: string) => {
  if (timers.has(label)) {
    console.warn(`Timer '${label}' already exists.`);
    return;
  }
  timers.set(label, {
    start: performance.now(),
    spinner: new Spinner({
      text: label,
    }),
  });
};
console.timeEnd = (label: string) => {
  if (!timers.has(label)) {
    console.warn(`Timer '${label}' does not exist.`);
    return;
  }
  const { start, spinner } = timers.get(label)!;
  timers.delete(label);
  spinner.success(label + chalk.blackBright(` [${(performance.now() - start).toFixed(3)}ms]`));
};
console.timeLog = (label: string) => {
  if (!timers.has(label)) {
    console.warn(`Timer '${label}' does not exist.`);
    return;
  }
  const { start } = timers.get(label)!;
  console.log(label + chalk.blackBright(` [${(performance.now() - start).toFixed(3)}ms]`));
};

const counters = new Map<string, number>();
console.count = (label: string) => {
  const n = (counters.get(label) || 0) + 1;
  counters.set(label, n);
  console.log(`${label}: ${n}`);
};
console.countReset = (label: string) => {
  counters.set(label, 0);
};

console.trace = trace;

process.on('uncaughtException', (exception: any) => {
  errorAllWidgets('uncaught exception');

  error(exception);

  writeLine('The above error was not caught by a catch block, execution cannot continue.');

  process.exit(1);
});
process.on('unhandledRejection', (reason: any) => {
  errorAllWidgets('unhandled rejection');
  error(reason);
  writeLine(
    '\nThe above error originated either by throwing inside of an async function without a catch block, or by rejecting a promise which was not handled with .catch()'
  );
  process.exit(1);
});