import chalk, { type ChalkInstance } from 'chalk';
import { writeSync } from 'fs';
import { inspect } from 'util';

export function convertHSVtoRGB(h: number, s: number, v: number): [number, number, number] {
    let r, g, b;
    const i = Math.floor(h * 6);
    const f = h * 6 - i;
    const p = v * (1 - s);
    const q = v * (1 - f * s);
    const t = v * (1 - (1 - f) * s);
    switch (i % 6) {
        case 0:
            (r = v), (g = t), (b = p);
            break;
        case 1:
            (r = q), (g = v), (b = p);
            break;
        case 2:
            (r = p), (g = v), (b = t);
            break;
        case 3:
            (r = p), (g = q), (b = v);
            break;
        case 4:
            (r = t), (g = p), (b = v);
            break;
        case 5:
            (r = v), (g = p), (b = q);
            break;
        default:
            return [0, 0, 0];
    }
    return [Math.round(r * 255), Math.round(g * 255), Math.round(b * 255)];
}

/** Converts non string objects into a string the way Node.js' console.log does it. */
export function stringify(...data: any[]) {
    return data.map(obj => (typeof obj === 'string' ? obj : inspect(obj, false, 4, true))).join(' ');
}

let buffer = '';
let exiting = false;

export function writeToStderr(data: string) {
    buffer += data;
    if (exiting) flushStderr();
}

export function flushStderr() {
    if (buffer) {
        writeSync(2, buffer);
        buffer = '';
    }
}

process.on('exit', () => {
    exiting = true;
    flushStderr();
});

export type CustomLoggerColor =
    | 'black'
    | 'red'
    | 'green'
    | 'yellow'
    | 'blue'
    | 'magenta'
    | 'cyan'
    | 'white'
    | 'gray'
    | 'grey'
    | 'blackBright'
    | 'redBright'
    | 'greenBright'
    | 'yellowBright'
    | 'blueBright'
    | 'magentaBright'
    | 'cyanBright'
    | 'whiteBright'
    | `#${string}`
    | number
    | [number, number, number];

export interface CustomLoggerOptions {
    id?: string;
    color?: CustomLoggerColor;
    coloredText?: boolean;
    boldText?: boolean;
    level?: number;
    debug?: boolean;
}

/** Matches `string`, `number`, and other objects with a `.toString()` method. */
export interface StringLike {
    toString(): string;
}

export interface FormatStringArgs {
    '%s': StringLike | null | undefined;
    '%d': number | null | undefined;
    '%i': number | null | undefined;
    '%f': number | null | undefined;
    '%x': number | null | undefined;
    '%X': number | null | undefined;
    '%o': any;
    '%O': any;
    '%c': string | null | undefined;
    '%j': any;
}

export type ProcessFormatString<S> = S extends `${string}%${infer K}${infer B}`
    ? `%${K}` extends keyof FormatStringArgs
    ? [FormatStringArgs[`%${K}`], ...ProcessFormatString<B>]
    : ProcessFormatString<B>
    : [];

export type LogData = string | number | boolean | object | null | undefined;

export interface LogFunction {
    /**
     * Writes data to the log. The first argument can be a printf-style format string, or usage
     * similar to `console.log`. Handles formatting objects including `Error` objects with pretty
     * colorized stack traces.
     *
     * List of formatters:
     *
     * - %s - String.
     * - %d, %f - Number.
     * - %i - Integer.
     * - %x - Hex.
     * - %X - Hex (uppercase)
     * - %o - Object.
     * - %O - Object (pretty printed).
     * - %j - JSON.
     */
    <S extends LogData>(data?: S, ...a: ProcessFormatString<S>): void;
    /** Calling a logger function with no arguments prints a blank line. */
    (): void;

    visible: boolean;
    name: string;
}

export function getColor(color: CustomLoggerColor): ChalkInstance {
    if (typeof color === 'string') {
        if (color in chalk) {
            return (chalk as any)[color];
        } else if (color.startsWith('#') || color.match(/^#[0-9a-fA-F]{6}$/)) {
            return chalk.hex(color);
        }
        throw new Error(`Invalid color: ${color}`);
    } else if (Array.isArray(color)) {
        return chalk.rgb(color[0], color[1], color[2]);
    }
    return chalk.ansi256(color);
}